C------------------------------------------------------------------------------
C       Example parallel program to show the use of the "pzssmp" routine 
c       in the peer mode.
C------------------------------------------------------------------------------
C.. This program can be downloaded via anonymous FTP from the following site:
C
C   ftp://ftp.cs.umn.edu/users/kumar/anshul/pzssmp_ex3.f
C
C   (C) IBM Corporation, 2002.
C
c	Acceptance and use of this program constitutes the user's understanding
c	that he/she will have no recourse to IBM for any actual or consequential
c	damages, including, but not limited to, lost profits or savings, arising
c	out of the use or inability to use this program. 
C------------------------------------------------------------------------------

        program pzssmp_ex3
        implicit none
        include 'mpif.h'
        integer mperr, myid, pp

        call wsetmaxthrds(1)
        call mpi_init(mperr)
        if (mperr .ne. 0) then
          print *,'MPI initialization error'
          stop
        end if
        call mpi_comm_rank(MPI_COMM_WORLD,myid,mperr)
        call mpi_comm_size(MPI_COMM_WORLD,pp,mperr)

        if (myid .eq. 0) then
          call master (pp)
	else
          call slave (myid)
        end if

        call mpi_finalize(mperr)
        stop
        end
C------------------------------------------------------------------------------
        subroutine master (pp)
        implicit none
        include 'mpif.h'
        integer nn, n, nrhs, mpierr, pp
        integer ia(*), ja(*), ibuf(*)
        real*8 avals(*), b(*), dbuf(*)
        pointer (pia,ia), (pja,ja), (pibuf,ibuf)
        pointer (pavals, avals), (pb,b), (pdbuf,dbuf)
        integer sml(8)
        integer i, myid, nnz, j, k, lnnz, partner, base, newn, oldn
        double precision waltime, rtc

        open(unit=91,file='/work1/agupta/21M/aa',status='unknown')
        open(unit=92,file='/work1/agupta/21M/b',status='unknown')
        open(unit=93,file='/work1/agupta/21M/ia',status='unknown')
        open(unit=94,file='/work1/agupta/21M/ja',status='unknown')

c       open(unit=91,file='/home3/anshul/bbn/toy/aa',status='unknown')
c       open(unit=92,file='/home3/anshul/bbn/toy/b',status='unknown')
c       open(unit=93,file='/home3/anshul/bbn/toy/ia',status='unknown')
c       open(unit=94,file='/home3/anshul/bbn/toy/ja',status='unknown')

        myid = 0
        read(92, *) nn, nrhs
        call wscmal(nn+1,i,pia)
        if (i .ne. 0) then
	  print *,myid,': IA malloc failed'
	  return
	end if
        do i = 1, nn + 1
          read(93, *) ia(i)
        end do
        close(93)
        nnz = ia(nn+1) - 1
        print *, 'n, nrhs, nnz = ',nn, nrhs, nnz

        j = nnz / pp
        k = 1
        base = 1
        do while (ia(k)-1 .lt. j .and. k .lt. nn)
	  k = k + 1
	end do
        n = k
        lnnz = ia(n+1) - 1

        call wsdmal(lnnz+lnnz,i,pavals)
        if (i .ne. 0) then
          print *,myid,': AVALS malloc failed'
          return
        end if
        call wscmal(lnnz,i,pja)
        if (i .ne. 0) then
          print *,myid,': JA malloc failed'
          return
        end if
        call wsdmal(n*nrhs*2,i,pb)
        if (i .ne. 0) then
          print *,myid,': B malloc failed'
          return
        end if

        waltime = rtc()
        do i = 1, n*nrhs
	  read(92, *) b(i+i-1), b(i+1)
	end do
        do i = 1, lnnz
          read(94, *) ja(i)
          read(91, *) avals(i+i-1), avals(i+i)
        end do
        print *,'Data read for 0 in time ',rtc()-waltime

        do partner = 1, pp - 1
	  waltime = rtc()
          oldn = k
	  k = k + 1
	  base = ia(k)
          do while (ia(k)-base .lt. j .and. k .lt. nn)
	    k = k + 1
	  end do
          newn = k - oldn
          lnnz = ia(k+1) - base
          sml(1) = nn
	  sml(2) = nrhs
	  sml(3) = newn
	  sml(4) = lnnz
	  call mpi_send(sml,4,MPI_INTEGER,partner,10,
     +                  MPI_COMM_WORLD,mpierr)
	  call mpi_send(ia(oldn+1),newn+1,MPI_INTEGER,partner,11,
     +                  MPI_COMM_WORLD,mpierr)
          call wsdmal(max(newn*nrhs,lnnz)*2,i,pdbuf)
          if (i .ne. 0) then
	    print *,'pdbuf alloc failed for',partner,'of size',lnnz
	    return
	  end if
          call wscmal(lnnz,i,pibuf)
          if (i .ne. 0) then
	    print *,'pibuf alloc failed for',partner,'of size',lnnz
	    return
	  end if
	  do i = 1, newn*nrhs
	    read(92, *) dbuf(i+i-1), dbuf(i+1)
	  end do
          call mpi_send(dbuf,newn*nrhs*2,MPI_DOUBLE_PRECISION,partner,
     +                  12,MPI_COMM_WORLD,mpierr)
          do i = 1, lnnz
	    read(94, *) ibuf(i)
	    read(91, *) dbuf(i+i-1), dbuf(i+i)
          end do
	  call mpi_send(ibuf,lnnz,MPI_INTEGER,partner,13,
     +                  MPI_COMM_WORLD,mpierr)
	  call mpi_send(dbuf,lnnz*2,MPI_DOUBLE_PRECISION,partner,14,
     +                  MPI_COMM_WORLD,mpierr)
	  call wscfr(pibuf)
	  call wsdfr(pdbuf)
          print *,'Data read & sent for',partner,'in time',
     +             rtc()-waltime
  	end do 

        close(92)
	close(91)
	close(94)
        call main_rtn_s(0,nn,n,nrhs,ia,ja,avals,b)

        return
        end
C------------------------------------------------------------------------------
        subroutine slave (myid)
        implicit none
        include 'mpif.h'
        integer myid, nn, n, nrhs, mpierr
        integer ia(*), ja(*)
        real*8 avals(*), b(*)
        pointer (pia,ia), (pja,ja)
        pointer (pavals, avals), (pb,b)
        integer sml(8), stat(MPI_STATUS_SIZE)
        integer i, j, lnnz

        call mpi_recv(sml,4,MPI_INTEGER,0,10,MPI_COMM_WORLD,
     +               stat,mpierr)
        nn = sml(1)
        nrhs = sml(2)
	n = sml(3)
	lnnz = sml(4)

        call wscmal(n+1,i,pia)
        if (i .ne. 0) then
          print *,myid,': IA malloc failed'
          return
        end if
	call mpi_recv(ia,n+1,MPI_INTEGER,0,11,MPI_COMM_WORLD,
     +                stat,mpierr)
        j = ia(1) - 1
        do i = 1, n + 1
	  ia(i) = ia(i) - j
	end do
	if (lnnz .ne. ia(n+1)-1) then
	  print *,myid,': ERR**: n, lnnz, ia(n+1) = ',n,lnnz,ia(n+1)
	  return
	end if
        call wscmal(lnnz,i,pja)
        if (i .ne. 0) then
          print *,myid,': JA malloc failed'
          return
        end if
        call wsdmal(lnnz+lnnz,i,pavals)
        if (i .ne. 0) then
          print *,myid,': AVALS malloc failed'
          return
        end if
        call wsdmal(n*nrhs*2,i,pb)
        if (i .ne. 0) then
          print *,myid,': B malloc failed'
          return
        end if

        call mpi_recv(b,n*nrhs*2,MPI_DOUBLE_PRECISION,0,12,
     +                MPI_COMM_WORLD,stat,mpierr)
	call mpi_recv(ja,lnnz,MPI_INTEGER,0,13,MPI_COMM_WORLD,
     +                stat,mpierr)
	call mpi_recv(avals,lnnz*2,MPI_DOUBLE_PRECISION,0,14,
     +                MPI_COMM_WORLD,stat,mpierr)
	
        call main_rtn_s(myid,nn,n,nrhs,ia,ja,avals,b) 

        return
        end
C------------------------------------------------------------------------------
        subroutine main_rtn_s(myid,nn,n,nrhs,ia,ja,avals,b)
        implicit none
        include 'mpif.h'
        integer myid,nn, n, ldb, nrhs, naux
        integer ia(n+1), ja(*)
        complex*16       avals(*), b(*)
        integer iparm(64)
        double precision dparm(64)
        integer mrp                     ! just a placeholder in this program
        complex*16       aux, diag      ! just placeholders in this program
        integer i, j
        double precision waltime, rtc, sanity
        integer perm(*), invp(*)
        pointer (pperm,perm), (pinvp,invp)

        call wscmal(nn,i,pperm)
        call wscmal(nn,j,pinvp)
        if (i+j .ne. 0) then
	  print *,myid,': malloc failed for perm, invp'
	  return
	end if

        print *,myid,': n, lnnz = ',n,ia(n+1)-1
	call mpi_barrier(MPI_COMM_WORLD,i)

        waltime = rtc()
        ldb = n
        naux = 0

C.. Fill 'iparm' and 'dparm' arrays with default values.

        iparm(1) = 0
        iparm(2) = 0
        iparm(3) = 0
        call pzssmp (n, ia, ja, avals, diag, perm, invp, b, ldb, nrhs,
     +               aux, naux, mrp, iparm, dparm)
        if (myid .eq. 0)
     +    print *,'Initialization complete in time - ',rtc()-waltime
        if (iparm(64) .ne. 0) then
          print *,myid,':The following ERROR was detected: ',iparm(64)
          return
        end if


C.. Ordering.

        waltime = rtc()
        iparm(2) = 1
        iparm(3) = 1
        call pzssmp (n, ia, ja, avals, diag, perm, invp, b, ldb, nrhs,
     +               aux, naux, mrp, iparm, dparm)
        if (myid .eq. 0)
     +    print *,'Ordering complete in time - ',rtc()-waltime
        if (iparm(64) .ne. 0) then
          print *,myid,':The following ERROR was detected: ',iparm(64)
          return
        end if

C.. Symbolic Factorization.

        waltime = rtc()
        iparm(2) = 2
        iparm(3) = 2
        call pzssmp (n, ia, ja, avals, diag, perm, invp, b, ldb, nrhs,
     +               aux, naux, mrp, iparm, dparm)
        if (myid .eq. 0)
     +    print *,'Symbolic complete in time - ',rtc()-waltime
        if (iparm(64) .ne. 0) then
          print *,myid,':The following ERROR was detected: ',iparm(64)
          return
        end if
        if (myid .eq. 0)
     +    print *,'Number of nonzeros in factor L = ',iparm(24)
        if (myid .eq. 0)
     +    print *,'Number of FLOPS in factorization = ',dparm(23)
        print *,myid,':Double words needed to factor =',iparm(23)

C.. Solve the same system using LDL' factorization.

        waltime = rtc()
        iparm(2) = 3
        iparm(3) = 3
        iparm(31) = 1
        call pzssmp (n, ia, ja, avals, diag, perm, invp, b, ldb, nrhs,
     +               aux, naux, mrp, iparm, dparm)
        waltime = rtc()-waltime
        if (myid .eq. 0)
     +    print *,'LDL^T factorization complete in time - ',waltime
        if (myid .eq. 0)
     +    print *,'Factorization MegaFlops = ',
     +            (dparm(23) * 1.d-6) / waltime
        if (iparm(64) .ne. 0) then
          print *,myid,':The following ERROR was detected: ',iparm(64)
          return
        end if

C.. Back substitution.

        waltime = rtc()
        iparm(2) = 4
        iparm(3) = 4
        call pzssmp (n, ia, ja, avals, diag, perm, invp, b, ldb, nrhs,
     +               aux, naux, mrp, iparm, dparm)
        if (myid .eq. 0)
     +    print *,'Back substitution complete in time - ',rtc()-waltime
        if (myid .eq. 0)
     +    print *,'Relative residual = ',dparm(7)
        if (iparm(64) .ne. 0) then
          print *,myid,':The following ERROR was detected: ',iparm(64)
          return
        end if

        sanity = 0.d0
        do i = 1, n * nrhs
          sanity = sanity + b(i)
        end do
        print *,myid,': Sanity = ',sanity

        return
        end
C------------------------------------------------------------------------------

