C------------------------------------------------------------------------------
C       Example program to show the use of the "wismp" routine for
c       solving a symmetric positive definite system using the default
c       preconditioner.
C------------------------------------------------------------------------------
C.. This program can be obtained from:
C
C   http://www.cs.umn.edu/~agupta/wsmp
C
C   (C) IBM Corporation, 2012.
C
c	Acceptance and use of this program constitutes the user's understanding
c	that he/she will have no recourse to IBM for any actual or consequential
c	damages, including, but not limited to, lost profits or savings, arising
c	out of the use or inability to use this program. 
C------------------------------------------------------------------------------

        program wssmp_ex1
        implicit none

        integer n, nnz, ldb, ldx, nrhs
        parameter (n = 99999, nnz = n)
        integer iparm(64)
        integer ia(n+1)
        integer ja(nnz)
        double precision dparm(64)
        double precision avals(nnz)
        double precision b(n), x(n), y(n)
        double precision derr
        double precision rmisc, cvgh      ! just placeholders in this program

        integer i
        double precision waltime, wsmprtc

C.. Fill all arrays containing matrix data.
  
        ldb = n
        ldx = n
        nrhs = 1 

        ia(1) = 1
        do i = 2,n
          ia(i) = ia(i-1) + (nnz/n)
        enddo
        ia(n+1) = nnz+1

        do i = 1,nnz
          ja(i) = i
          avals(i) = 1.0
        enddo

C.. Executable part of the program starts here.
  
        waltime = wsmprtc()

C.. Fill 'iparm' and 'dparm' arrays with default values.

	call wsetmaxthrds(2)	! Use 1 thread
        iparm(1) = 0
        iparm(2) = 0
        iparm(3) = 0
        call wismp (n, ia, ja, avals, b, ldb, x, ldx, nrhs,
     +              rmisc, cvgh, iparm, dparm)

        print *,'Initialization complete in time - ',wsmprtc()-waltime
        if (iparm(64) .ne. 0) then
          print *,'The following ERROR was detected: ',iparm(64)
          stop
        end if

C.. Analyze structure.

        waltime = wsmprtc()
        iparm(2) = 1
        iparm(3) = 1
        call wismp (n, ia, ja, avals, b, ldb, x, ldx, nrhs,
     +              rmisc, cvgh, iparm, dparm)
        if (iparm(64) .ne. 0) then
          print *,'The following ERROR was detected: ',iparm(64)
          stop
        end if
        print *,'Step 1 complete in time - ',wsmprtc()-waltime

C.. Analyze Values.

        waltime = wsmprtc()
        iparm(2) = 2
        iparm(3) = 2
        call wismp (n, ia, ja, avals, b, ldb, x, ldx, nrhs,
     +              rmisc, cvgh, iparm, dparm)
        if (iparm(64) .ne. 0) then
          print *,'The following ERROR was detected: ',iparm(64)
          stop
        end if
        print *,'Step 2 complete in time - ',wsmprtc()-waltime

C.. Matrix-vector multiplication (compute b = Ax).

        do i = 1, n
          x(i) = 1.d0 + 1.d-1 * dfloat(i-(i/10)*10)
        end do

        waltime = wsmprtc()
        iparm(2) = 5
        iparm(3) = 5
        call wismp (n, ia, ja, avals, b, ldb, x, ldx, nrhs,
     +              rmisc, cvgh, iparm, dparm)
        if (iparm(64) .ne. 0) then
          print *,'The following ERROR was detected: ',iparm(64)
          stop
        end if
        print *,'Matrix.vector done in time - ',wsmprtc()-waltime

        stop
        end

