/*----------------------------------------------------------------------------*/
/*	(C) Copyright IBM Corporation 2012.                                   */
/*----------------------------------------------------------------------------*/
//       Example program to show the use of the "zismp" routine for
//       solving a complex usymmetric system using the SSOR preconditioner.

#include <stdio.h>
#include <stdlib.h>

extern void zismp(int *, int *, int *, double *,  double *, int *, double *, 
                  int *, int *, double *, double *, int *, double *);
extern double wsmprtc();

int main(int argc,char **argv)
{
        int n, ldb, ldx, ldy, nrhs;
        int iparm[64];
        double dparm[64];
        double b[16], y[16];	// declared double, but actually complex

        int i, nthd;
        double derr, rmisc, tt1, tt2;

//C.. Fill all arrays containing matrix data.

        n = 8; ldb = 8; ldx = 8; ldy = 8; nrhs = 1;
        int ia[] = {0, 4, 8, 13, 18, 22, 28, 34, 42};

        int ja[]  = {
     0,          2,                      6,    7,
     0,    1,    2,                            7,
     0,    1,    2,                      6,    7,
     0, 	       3,          5,    6,    7,
     	   1,                4,    5,          7,
     		 2,    3,    4,    5,    6,    7,
     0,          2,    3,          5,    6,    7,
     0,    1,    2,    3,    4,    5,    6,    7,
		 };

     double avals[] = { 	// declared double, but actually complex
8.5,1.0,          1.0,0.0,                            1.0,3.0, 3.0,0.0,
3.5,0.0, 9.5,2.0, 1.0,0.5,                                     3.0,1.0,
1.0,2.0, 1.0,0.5, 6.0,2.0,                            2.0,4.0, 4.0,2.0,
8.0,0.5,                   8.0,9.0,          1.0,0.0, 1.0,3.0, 3.0,0.0,
         7.0,3.0,                   9.0,8.0, 1.0,0.0,          3.0,1.0,
                  6.0,3.0, 1.0,0.0, 1.0,2.0, 6.0,2.0, 2.0,4.0, 4.0,2.0,
1.0,0.0,          2.0,1.0, 1.0,0.5,          2.0,0.0, 6.0,4.0, 4.0,0.0,
3.0,2.0, 3.0,3.0, 4.0,4.0, 3.0,2.0, 3.0,3.0, 4.0,1.0, 4.0,4.0, 7.0,4.0,
		      };

      double x[] = {1.1,1.8, 1.2,1.7, 1.3,1.6, 1.4,1.5, 
                    1.5,1.4, 1.6,1.3, 1.7,1.2, 1.8,1.1};

//C.. Executable part of the program starts here.

//C.. Fill 'iparm' and 'dparm' arrays with default values.

        nthd = 2;	// 		! Use 2 threads
        wsetmaxthrds_(&nthd);

        iparm[0] = 0;
        iparm[1] = 0;
        iparm[2] = 0;
        zismp_(&n, ia, ja, avals, b, &ldb, x, &ldx, &nrhs, &rmisc, NULL, iparm, 
               dparm);
        if (iparm[63] != 0) {
          printf("The following ERROR was detected in S0: %d", iparm[63]);
          exit(1);
        }

//C.. Modify some defaults

	iparm[4] = 0;  //	 	! C-style numbering
	iparm[14] = 3;  //	 	! ILU preconditioning

//C.. Analyze structure

	tt1 = wsmprtc();
        iparm[1] = 1;
        iparm[2] = 1;
        zismp_(&n, ia, ja, avals, b, &ldb, x, &ldx, &nrhs, &rmisc, NULL, iparm, 
               dparm);
        if (iparm[63] != 0) {
          printf("The following ERROR was detected in S1: %d",iparm[63]);
          exit(1);
        }
	tt2 = wsmprtc();
        printf("Step 1 complete in time - %g\n",tt2-tt1);

//C.. Analyze Values

        tt1 = wsmprtc();
        iparm[1] = 2;
        iparm[2] = 2;
        zismp_(&n, ia, ja, avals, b, &ldb, x, &ldx, &nrhs, &rmisc, NULL, iparm, 
               dparm);
        if (iparm[63] != 0) {
          printf("The following ERROR was detected in S2: %d",iparm[63]);
          exit(1);
        }
        tt2 = wsmprtc();
        printf("Step 2 complete in time - %g\n",tt2-tt1);

//C.. Matrix-Vector multiplication (compute b = Ax).

        tt1 = wsmprtc();
        iparm[1] = 5;
        iparm[2] = 5;
        zismp_(&n, ia, ja, avals, b, &ldb, x, &ldx, &nrhs, &rmisc, NULL, iparm, 
               dparm);
        if (iparm[63] != 0) {
          printf("The following ERROR was detected in S5: %d", iparm[63]);
          exit(1);
        }
        tt2 = wsmprtc();
        printf("Step 5 complete in time - %g\n",tt2-tt1);

//C.. Generate Preconditioner.

        tt1 = wsmprtc();
        iparm[1] = 3;
        iparm[2] = 3;
        zismp_(&n, ia, ja, avals, b, &ldb, x, &ldx, &nrhs, &rmisc, NULL, iparm,
               dparm);
        if (iparm[63] != 0) {
          printf("The following ERROR was detected in S3: %d", iparm[63]);
          exit(1);
        }
        tt2 = wsmprtc();
        printf("Step 3 complete in time - %g\n",tt2-tt1);

//C.. Iterative Solution (solve for y in Ay = b).

        tt1 = wsmprtc();
        iparm[1] = 4;
        iparm[2] = 4;
        zismp_(&n, ia, ja, avals, b, &ldb, y, &ldy, &nrhs, &rmisc, NULL, iparm,
               dparm);
        if (iparm[63] != 0) {
          printf("The following ERROR was detected in S4: %d", iparm[63]);
          exit(1);
        }
        tt2 = wsmprtc();
        printf("Step 4 complete in %g seconds in %d iters\n",tt2-tt1,iparm[25]);

	rmisc = 0.0;
	derr = 0.0;
        for (i = 0; i < n; i++) {
	  printf("%d : (%g, %g)\n",i,y[i+i],y[i+i+1]);
	  derr = derr + (y[i+i] - x[i+i]) * (y[i+i] - x[i+i]);
          rmisc = rmisc + b[i+i] * b[i+i];
	  derr = derr + (y[i+i+1] - x[i+i+1]) * (y[i+i+1] - x[i+i+1]);
          rmisc = rmisc + b[i+i+1] * b[i+i+1];
	}
        printf("Residual = %g; Relative error = %g\n",dparm[25],derr/rmisc);

	return 0;
}

