C------------------------------------------------------------------------------
C       Example program to show the use of the "zssmp" routine for
c	IPARM(31) = 3 and 4.
C------------------------------------------------------------------------------
C.. This program can be obtained from:
C
C   http://www.cs.umn.edu/~agupta/wsmp
C
C   (C) Copyright IBM Corporation, 2002, 2015.
C
c	Acceptance and use of this program constitutes the user's understanding
c	that he/she will have no recourse to IBM for any actual or consequential
c	damages, including, but not limited to, lost profits or savings, arising
c	out of the use or inability to use this program. 
C------------------------------------------------------------------------------

        program zssmp_ex2
        implicit none

        integer n, ldb, nrhs, naux
        integer iparm(64)
        integer ia(10)
        integer ja(29)
        integer perm(9), invp(9)
        double precision dparm(64), err
        complex*16       avals(29)
        complex*16       b(9), x(9), y(9), v, w

        integer mrp                     ! just a placeholder in this program
        double precision aux, diag      ! just placeholders in this program

        integer i, j, k
        double precision waltime, wsmprtc

C.. Fill all arrays containing matrix data.

        data n /9/, ldb /9/, nrhs /1/, naux /0/

        data ia /1,5,9,13,17,21,25,27,29,30/

        data ja
     1        /1,          3,                      7,    8,
     2               2,    3,                            8,    9,
     3                     3,                      7,    8,    9,
     4                           4,          6,    7,    8,
     5                                 5,    6,          8,    9,
     6                                       6,    7,    8,    9,
     7                                             7,    8,
     8                                                   8,    9,
     9                                                         9/
        data avals
     1/(14.d0,-14.d0), (-1.d0,0.d0), (-1.d0,0.d0), (-3.d0,0.d0),
     2 (14.d0,-14.d0), (-1.d0,0.d0), (-3.d0,0.d0), (-1.d0,0.d0),
     3 (16.d0,-16.d0), (-2.d0,0.d0), (-4.d0,0.d0), (-2.d0,0.d0),
     4 (14.d0,-14.d0), (-1.d0,0.d0), (-1.d0,0.d0), (-3.d0,0.d0),
     5 (14.d0,-14.d0), (-1.d0,0.d0), (-3.d0,0.d0), (-1.d0,0.d0),
     6 (16.d0,-16.d0), (-2.d0,0.d0), (-4.d0,0.d0), (-2.d0,0.d0),
     7 (16.d0,-16.d0), (-4.d0,0.d0),
     8 (71.d0,-71.d0), (-4.d0,0.d0),
     9 (16.d0,-16.d0)/

C.. Executable part of the program starts here.

        call wsetmaxthrds(2)    ! Use 2 threads

        waltime = wsmprtc()

C.. Fill 'iparm' and 'dparm' arrays with default values.

C.. As an alternative to this step, the values in 'iparm' and 'dparm' can be
c   filled with values suitable for the application either manually or by 
c   using a "data" statement according to the description in the User's guide.

        iparm(1) = 0
        iparm(2) = 0
        iparm(3) = 0
        call zssmp (n, ia, ja, avals, diag, perm, invp, b, ldb, nrhs,
     +              aux, naux, mrp, iparm, dparm)

        print *,'Initialization complete in time - ',wsmprtc()-waltime
        if (iparm(64) .ne. 0) then
          print *,'The following ERROR was detected: ',iparm(64)
          stop
        end if

C.. Ordering.

        waltime = wsmprtc()
        iparm(2) = 1
        iparm(3) = 1
	iparm(31) = 3
        call zssmp (n, ia, ja, avals, diag, perm, invp, b, ldb, nrhs,
     +              aux, naux, mrp, iparm, dparm)

        print *,'Ordering complete in time - ',wsmprtc()-waltime

        if (iparm(64) .ne. 0) then
          print *,'The following ERROR was detected: ',iparm(64)
          stop
        end if

C.. Symbolic Factorization.

        waltime = wsmprtc()
        iparm(2) = 2
        iparm(3) = 2
        call zssmp (n, ia, ja, avals, diag, perm, invp, b, ldb, nrhs,
     +              aux, naux, mrp, iparm, dparm)

        print *,'Symbolic complete in time - ',wsmprtc()-waltime
        if (iparm(64) .ne. 0) then
          print *,'The following ERROR was detected: ',iparm(64)
          stop
        end if

        print *,'Number of nonzeros in factor L = 1000 X ',iparm(24)
        print *,'Number of FLOPS in factorization = ',dparm(23)
        print *,'Double words needed to factor = 1000 X ',iparm(23) 

C.. LDL' Factorization with pivoting

        waltime = wsmprtc()
        iparm(2) = 3
        iparm(3) = 3
        iparm(25) = 1			! Condition number estimate
        call zssmp (n, ia, ja, avals, diag, perm, invp, b, ldb, nrhs,
     +              aux, naux, mrp, iparm, dparm)

        waltime = wsmprtc() - waltime
        print *,'Cholesky complete in time - ',waltime
        print *,'Factorization MegaFlops = ',
     +           (dparm(23) * 1.d-6) / waltime
        if (iparm(64) .ne. 0) then
          print *,'The following ERROR was detected: ',iparm(64)
          stop
        end if

C.. Back substitution.

        do i = 1, n
          b(i) = (1.d0,1.d0)
          x(i) = (1.d0,1.d0)
          y(i) = (0.d0,0.d0)
        end do

        waltime = wsmprtc()
        iparm(2) = 4
        iparm(3) = 4
        call zssmp (n, ia, ja, avals, diag, perm, invp, x, ldb, nrhs,
     +              aux, naux, mrp, iparm, dparm)

        print *,'Back substitution done in time - ',wsmprtc()-waltime       
        if (iparm(64) .ne. 0) then
          print *,'The following ERROR was detected: ',iparm(64)
          stop
        end if

C.. Iterative refinement.

        waltime = wsmprtc()
        iparm(2) = 5
        iparm(3) = 5
        call zssmp (n, ia, ja, avals, diag, perm, invp, x, ldb, nrhs,
     +              aux, naux, mrp, iparm, dparm)

        print *,'Iterative refinement done in time - ',wsmprtc()-waltime       
        if (iparm(64) .ne. 0) then
          print *,'The following ERROR was detected: ',iparm(64)
          stop
        end if

	do i = 1, n
          v = x(i)
	  w = y(i) + v * avals(ia(i))
	  do j = ia(i)+1, ia(i+1) - 1
	    k = ja(j)
	    w = w + x(k) * avals(j)
	    y(k) = y(k) + v * avals(j)
	  end do
	  y(i) = w
	end do

	err = 0.d0
	do i = 1, n
	  err = err + zabs(b(i) - y(i))
	  y(i) = (0.d0, 0.d0)
	end do
	err = err / dfloat(n)

        print *,'Rel, 1-norm of Err = ',dparm(7), err
        print *,'The solution of the system is as follows:'
        do i = 1, n
          print *,i,' : ',x(i)
        end do

C.. Solve the same system using LDL' factorization w/o pivoting.

C.. Ordering.

1000	continue

        waltime = wsmprtc()
        iparm(2) = 1
        iparm(3) = 1
        iparm(31) = 4
        call zssmp (n, ia, ja, avals, diag, perm, invp, b, ldb, nrhs,
     +              aux, naux, mrp, iparm, dparm)

        print *,'Ordering complete in time - ',wsmprtc()-waltime

        if (iparm(64) .ne. 0) then
          print *,'The following ERROR was detected: ',iparm(64)
          stop
        end if

C.. Symbolic Factorization.

        waltime = wsmprtc()
        iparm(2) = 2
        iparm(3) = 2
        call zssmp (n, ia, ja, avals, diag, perm, invp, b, ldb, nrhs,
     +              aux, naux, mrp, iparm, dparm)

        print *,'Symbolic complete in time - ',wsmprtc()-waltime
        if (iparm(64) .ne. 0) then
          print *,'The following ERROR was detected: ',iparm(64)
          stop
        end if

        print *,'Number of nonzeros in factor L = 1000 X ',iparm(24)
        print *,'Number of FLOPS in factorization = ',dparm(23)
        print *,'Double words needed to factor = 1000 X ',iparm(23) 

        waltime = wsmprtc()
        iparm(2) = 3
        iparm(3) = 3
        call zssmp (n, ia, ja, avals, diag, perm, invp, b, ldb, nrhs,
     +              aux, naux, mrp, iparm, dparm)

        waltime = wsmprtc()-waltime
        print *,'LDL^T factorization complete in time - ',waltime
        print *,'Factorization MegaFlops = ',
     +           (dparm(23) * 1.d-6) / waltime
        if (iparm(64) .ne. 0) then
          print *,'The following ERROR was detected: ',iparm(64)
          stop
        end if
        print *,'Condition number = ',dparm(2)*dparm(3)

C.. Back substitution.

        do i = 1, n
          x(i) = b(i)
          y(i) = (0.d0, 0.d0)
        end do

        waltime = wsmprtc()
        iparm(2) = 4
        iparm(3) = 4
        iparm(7) = 3
        call zssmp (n, ia, ja, avals, diag, perm, invp, x, ldb, nrhs,
     +              aux, naux, mrp, iparm, dparm)

        print *,'Back substitution done in time - ',wsmprtc()-waltime
        if (iparm(64) .ne. 0) then
          print *,'The following ERROR was detected: ',iparm(64)
          stop
        end if

	do i = 1, n
          v = x(i)
	  w = y(i) + v * avals(ia(i))
	  do j = ia(i)+1, ia(i+1) - 1
	    k = ja(j)
	    w = w + x(k) * avals(j)
	    y(k) = y(k) + v * avals(j)
	  end do
	  y(i) = w
	end do

	err = 0.d0
	do i = 1, n
	  err = err + zabs(b(i) - y(i))
	  y(i) = (0.d0, 0.d0)
	end do
	err = err / dfloat(n)

        print *,'Rel, 1-norm of Err = ',dparm(7), err

C.. Iterative refinement.

        waltime = wsmprtc()
        iparm(2) = 5
        iparm(3) = 5
        call zssmp (n, ia, ja, avals, diag, perm, invp, x, ldb, nrhs,
     +              aux, naux, mrp, iparm, dparm)

        print *,'Iterative refinement done in time - ',wsmprtc()-waltime       
        if (iparm(64) .ne. 0) then
          print *,'The following ERROR was detected: ',iparm(64)
          stop
        end if
        do i = 1, n
          v = x(i)
          w = y(i) + v * avals(ia(i))
          do j = ia(i)+1, ia(i+1) - 1
            k = ja(j)
            w = w + x(k) * avals(j)
            y(k) = y(k) + v * avals(j)
          end do
          y(i) = w
        end do

        err = 0.d0
        do i = 1, n
          err = err + zabs(b(i) - y(i))
        end do
        err = err / dfloat(n)

        print *,'Rel, 1-norm of Err = ',dparm(7), err

        print *,'The solution of the system is as follows:'
        do i = 1, n
          print *,i,' : ',x(i)
        end do

        stop
        end

